#include <stdio.h> 
#include <stdlib.h>
#include "unitesLexicales.h"
#include <string.h>


/*Variables globales*/
int uc;

/*Variables globales importes de l'analyseur lexical*/
extern int noligne;
extern int valeur;
extern float valeurf;
extern char *yytext;
extern FILE * yyin;

/*Prototypes des fonctions de l'analyseur syntaxique*/
void math_expr(void);
void id_list(void);
void liste_bornes(void);
void id_tableau(void);
void for_list(void);
void expr_list(void);
void list_arg_ext(void);
void arg_list(void);
void id_suivant(void);
void id_simple(void);
void proc(void);
void id_tableau(void);
void declaration(void);

void body(void);
void block(void);
void program(void);

/*Fonctions utilitaires*/
/*Fonction erreur:
Affiche un message sur la sortie standard et arrte le processus
Ncessite la variable importe noligne en global*/
void erreur(char* message){
	fprintf(stderr,"SYNTAX ERROR LINE %d: %s\n",noligne ,message);
	exit(EXIT_FAILURE);
}


/*Fonctions de l'analyseur syntaxique*/

/*Section expressions mathmatiques*/

void facteur(void){
  switch(uc){
  case BRACKETL:
    uc=yylex();
    math_expr();
    if(uc!=BRACKETR)
      erreur("missing ) after end of bracketted expression");
    uc=yylex();
    break;
  case NBR_INT:
  case NBR_REAL:
    uc=yylex();
    break;
  case IDENT:
    uc=yylex();
    if(uc==ARRAYBRACKETL){
      uc=yylex();
      math_expr();
      if(uc!=ARRAYBRACKETR)
	erreur("missing ] at the end of array indexation expression");
      uc=yylex();
      break;
    }
    else{
      if(uc==BRACKETL){//Appel  fonction
	uc=yylex();
	if(uc!=BRACKETR){
	  math_expr();
	  while(uc==COMMA){
	    uc=yylex();
	    math_expr();
	  }
	  if(uc!=BRACKETR)
	    erreur("missing ) at the end of function parameter in expression");
	  uc=yylex();
	}
	else{
	  //Si il y a BRACKETR on avance (fonction sans paramtres)
	  uc=yylex();
	}
      }
    }
    break;
    default:
      erreur("unauthorized operation attempted in math expression, aborting evaluation");
  }
}

void terme(void){
  facteur();
  while(uc==MULT || uc==DIV || uc==DIVENT || uc==POWER){
    uc=yylex();
    facteur();
  }
}

void expArith(void){
  terme();
  while(uc==PLUS || uc==MINUS){
    uc=yylex();
    terme();
  }
}

void comparaison(void){
  expArith();
  switch(uc){
  case EQUAL:
  case NOTEQUAL:
  case LESS:
  case NOTGREATER:
  case NOTLESS:
  case GREATER:
    uc=yylex();
    expArith();
  default:
    ;
  }
}

void conjonction(void){
  comparaison();
  while(uc==AND){
    uc=yylex();
    comparaison();
  }
}

void math_expr(void){
  if(uc==NOT){
    uc=yylex();
  }
  conjonction();
  while(uc==OR){
    uc=yylex();
    conjonction();
  }
}

/*Section Instructions*/

void for_type(void){
  if(uc==STEP){
    uc=yylex();
    math_expr();
    if(uc==UNTIL){
      uc=yylex();
      math_expr();
    }
    else
      erreur("missing 'UNTIL' keyword with 'STEP'-type 'FOR' statement");
  }
  else{
    if(uc==WHILE){
      uc=yylex();
      math_expr();
    }
  }
}

void for_list_element(void){
  math_expr();
  for_type();
}

void for_listc(void){
  switch(uc){
  case COMMA:
    uc=yylex();
    for_list();
    break;
  case DO:
    uc=yylex();
    break;
  default:
    erreur("abnormal termination for list in 'FOR' statement");
  }
}

void for_list(void){
  for_list_element();
  for_listc();
}

void for_clause(void){
  if(uc!=IDENT)
    erreur("missing identifier after 'FOR' statement");
  uc=yylex();
  if(uc!=AFFECT){
    erreur("missing := for identifier in 'FOR' statement");
  }
  uc=yylex();
  for_list();
}

void else_bloc(void){
  if(uc==ELSE){
    uc=yylex();
    body();
  }
}

void then_bloc(void){
  if(uc==THEN){
    uc=yylex();
    body();
  }
  else
    erreur("missing then before statement block");
}

void expr_suiv(){
  if(uc==COMMA){
    uc=yylex();
    expr_list();
  }
  else{
    if(uc==BRACKETR)
      uc=yylex();
    else
      erreur("missing , or ) in expression list");
  }
}

void expr_list(void){
  if(uc==OPENSTRING){
    uc=yylex();
  }
  else
    math_expr();
  expr_suiv();
}


void arg_suivant(void){
  if(uc==COMMA){
    uc=yylex();
    arg_list();
  }
  else{
    if(uc==BRACKETR){
      uc=yylex();
      if(uc!=COMMADOT)
	erreur("Missing ; after argument list");
      uc=yylex();
    }
  }
}

void arg_list(void){
  if(uc==BRACKETR){
    uc=yylex();
    if(uc!=COMMADOT)
      erreur("Missing ; after argument list");
    uc=yylex();
  }
  else{
    if(uc==OPENSTRING){
      uc=yylex();
    }
    else
      math_expr();
    arg_suivant();
  }
}

void voidproc_affect(void){
  if(uc==AFFECT){
    uc=yylex();
    math_expr();
    if(uc==COMMADOT)
      uc=yylex();
    else
      erreur("missing ; at the end of affectation statement");
  }
  else{
    if(uc==BRACKETL){
      uc=yylex();
      arg_list();
    }
    else
      erreur("Missing := or ( in identifier-beginning statement");
  }
}

void tab(void) {
  if(uc==ARRAYBRACKETL){
    uc=yylex();
    math_expr();
    if(uc!=ARRAYBRACKETR)
      erreur("missing ] at the end of table indexation expression");
    uc=yylex();
  }
}

void instr(void){
  //Liste des instructions, for, if et identifiants
  switch(uc){
  case FOR:
    uc=yylex();
    for_clause();
    body();
    break;
  case IF:
    uc=yylex();
    math_expr();
    then_bloc();
    else_bloc();
    break;
  case IDENT:
    uc=yylex();
    tab();
    voidproc_affect();
    break;
  default:
    erreur("unrecognized statement token");
  }
}

void instructions(void){
  if(uc!=BEND){
    instr();
    if(uc==COMMADOT){
      uc=yylex();
      instructions();
    }
    else
      erreur("missing ; in end of statement");
  }
  else{
    //Fin de bloc
    uc=yylex();
  }
}

/*Section declaration*/
void tab_proc(void){
  if(uc==PROCEDURE){
    uc=yylex();
  }
  else{
    if(uc==ARRAY){
      uc=yylex();
    }
  }
}

void specifier(void){
  switch(uc){
  case STRING:
  case ARRAY:
  case PROCEDURE:
    uc=yylex();
    break;
  case INTEGER:
  case REAL:
  case BOOLEAN:
    uc=yylex();
    tab_proc();
    break;
  default:
    erreur("invalid specifier");
  }
}

void spec(void){
  specifier();
  //id_simple();
  if(uc==IDENT){
    uc=yylex();
    while(uc==COMMA){
      uc=yylex();
      if(uc!=IDENT)
	erreur("Missing IDENTIFIER after , in list of values");
      uc=yylex();
      }
    if(uc!=COMMADOT)
      erreur("Missing ; after Value specifier");
    uc=yylex();
  }
  else{
    erreur("Missing IDENTIFIER after value statement");
  }
}

void specifications(void){
  switch(uc){
  case STRING:
  case ARRAY:
  case INTEGER:
  case REAL:
  case BOOLEAN:
  case PROCEDURE:
    spec();
    specifications();
    break;
  default:
    block();
    break;
  }
}

void proc_specifiers(void){
  if(uc==VALUE){
    uc=yylex();
    //id_simple();
    /*Reconnaitre liste d'identifiants*/
    if(uc==IDENT){
      uc=yylex();
      while(uc==COMMA){
	uc=yylex();
	if(uc!=IDENT)
	  erreur("Missing IDENTIFIER after , in list of values");
	uc=yylex();
      }
      if(uc!=COMMADOT)
	erreur("Missing ; after Value specifier");
      uc=yylex();
    }
    else{
      erreur("Missing IDENTIFIER after value statement");
    }
    specifications();
  }
  else{
    switch(uc){
    case STRING:
    case ARRAY:
    case INTEGER:
    case REAL:
    case BOOLEAN:
    case PROCEDURE:
      specifications();
      break;
    default:
      //Bloc
      block();
    }
  }
}

void id_suivantp(void){
  if(uc==COMMA){
    uc=yylex();
    id_list();
  }
  else{
    if(uc!=BRACKETR)
      erreur("missing ) at the end of parameters list in procedure declaration");
    uc=yylex();
  }
}

void id_list(void){
  if(uc==IDENT){
    uc=yylex();
    id_suivantp();
  }
  else
    erreur("missing IDENTIFIER in list");
}

void param_externe(void){
  if(uc==IDENT){
    uc=yylex();
    if(uc==DOUBLEDOT){
      uc=yylex();
      if(uc==BRACKETL){
	uc=yylex();
	id_list();
      }
    }
    else
      erreur("Missing : for external parameter");
  }
  else
    erreur("CHAIN OF CHARS missing for external parameter");
}

void param_externes(void){
  while(uc!=COMMADOT){
    param_externe();
  }
  uc=yylex();
}

void id_list_vide(void){
  if(uc!=BRACKETR)
    id_list();
  else
    uc=yylex();
}

void proc_head(void){
  if(uc!=BRACKETL)
    erreur("missing ( after procedure identifier");
  uc=yylex();
  id_list_vide();
  if(uc!=COMMADOT)
    erreur("Missing ; after header of procedure declaration");
  uc=yylex();
  //param_externes();
}

void id_suivant(void){
  if(uc==COMMA){
    uc=yylex();
    id_simple();
  }
  else{
    if(uc!=COMMADOT)
      erreur("Termination ; missing after simple identifier declaration");
    uc=yylex();
  }
}

void id_simple(void){
  if (uc==IDENT){
    uc=yylex();
    id_suivant();
  }
}

void paire(void){
  if(uc==COMMA){
    uc=yylex();
    liste_bornes();
  }
  else{
    if(uc!=ARRAYBRACKETR)
      erreur("missing ] at the end of extrema");
    uc=yylex();
  }
}

void paire_bornes(void){
  if(uc!=NBR_INT)
    erreur("Missing constant integer value for table minimum");
  uc=yylex();
  if(uc!=DOUBLEDOT)
    erreur("missing : for separator between table extrems");
  uc=yylex();
  if(uc!=NBR_INT)
    erreur("Missing constant integer value for table maximum");
  uc=yylex();
}

void liste_bornes(void){
  paire_bornes();
  paire();
}

void tableau_suivant(void){
  if(uc==COMMA){
    uc=yylex();
    id_tableau();
  }
  else {
    if(uc!=COMMADOT)
      erreur("Termination ; missing at the end of table declaration");
    uc=yylex();
  }
}

void tableau(void){
  if(uc==ARRAY){
    uc=yylex();
    id_tableau();
  }
  else{
    id_simple();
  }
}

void tableau_proc(void){
  switch(uc){
  case PROCEDURE:
    uc=yylex();
    proc();
    break;
  case ARRAY:
    uc=yylex();
    id_tableau();
    break;
  default:
    id_simple();
  }
}

void type_var(void){
  switch(uc){
  case ARRAY:
    uc=yylex();
    id_tableau();
    break;
  case INTEGER:
  case BOOLEAN:
  case REAL:
    uc=yylex();
    tableau();
    break;
  default:
    erreur("unrecognized variable type");
  }
}

void proc(void){
  if(uc!=IDENT)
    erreur("missing correct identifier for procedure declaration");
  uc=yylex();
  proc_head();
  proc_specifiers();
}

void id_tableau(void) {
  if(uc != IDENT)
    erreur("missing IDENTIFIER for array declaration");
  uc=yylex();
  if(uc != ARRAYBRACKETL)
    erreur("missing [ after IDENTIFIER in array declaration");
  uc=yylex();
  liste_bornes();
  tableau_suivant();
}

void declaration(void) {
  //Une seule dclaration pouvant concerner plusieurs variables
  //Spcialisation par type dclar
  switch(uc){
  case ARRAY:
    uc=yylex();
    id_tableau();
    break;
  case INTEGER:
  case REAL:
  case BOOLEAN:
    uc=yylex();
    tableau_proc();
    break;
  case OWN:
    uc=yylex();
    type_var();
    break;
  case PROCEDURE:
    uc=yylex();
    proc();
    break;
  }
}

void instr_unique(){
  instr();
}

void block(void) {
  //On attend un BEGIN
  if(uc != BBEGIN)
    erreur("missing 'BEGIN' at beginning of block!");
  //On avance
  uc = yylex();
  //Ensemble de Declarations optionnel
  while(uc==INTEGER || uc==OWN || uc==REAL || uc==BOOLEAN || uc==PROCEDURE || uc==ARRAY) {
    declaration();
  }
  //Ensemble d'instructions optionnelles
  while(uc==IDENT || uc == FOR || uc == IF) {
    instr();
  }
  if(uc!=BEND)
    erreur("missing 'END' at the end of block!");
  uc=yylex();
}

void body(void){
  if(uc==BBEGIN){
    block();
  }
  else {
    instr_unique();
  }
}

void program(void) {
  body();
}

int main(int argc, char * argv[]){
  int i;
  FILE *f;
  char* nomfic;
  if(argc != 2+1){
    fprintf(stderr, "USAGE: %s source_file.alg out_filename",argv[0]);
    exit(EXIT_FAILURE);
  }
  yyin = fopen(argv[1], "r");
  uc = yylex();
  program();
  printf("ALGOL60 syntax was successfully recognized\n");
  return EXIT_SUCCESS;
}
