#include <stdio.h>
#include "dictionnaire.h"
#include <assert.h>
#include <string.h>
#include <malloc.h>

static void killContent(DicElem *courant);

static DicElem *Global;
static DicElem *Local;
static DicElem *Argument;
static DicElem *CurGlobal;
static DicElem *CurLocal;
static DicElem *CurArgument;
static int idCount=0;

static DICContext Context=DIC_GLOBAL;
static SYMType BuilderType;
static SYMClass BuilderClass;
static SYMClass BuilderRef=ARG_VALUE;
static int BuilderOffset=0;

//choisit le contexte courant
void dicSetContext(DICContext context)
{
    Context = context;
}

//renvoie le contexte courant
DICContext dicGetContext()
{
    return Context;
}

//definit le type et la classe de la prochaine entree pour le constructeur
void dicSetBuilderType(SYMType type)
{
    BuilderType = type;
}

void dicSetBuilderClass(SYMClass classe)
{
    BuilderClass = classe;
}

void dicSetBuilderOffset(int offset)
{
    BuilderOffset = offset;
}
void dicSetBuilderRef(ARGType ref)
{
    BuilderRef = ref;
}

void dicSetType(SYMType type)
{
    assert(Context != DIC_ARGUMENTS);
    switch (Context)
    {
    case DIC_GLOBAL:
        assert(CurGlobal != NULL);
        CurGlobal->type = type;
        break;
    case DIC_LOCAL:
        assert(CurLocal != NULL);
        CurLocal->type = type;
        break;
    }
}

void dicSetClass(SYMClass classe)
{
    assert(Context != DIC_ARGUMENTS);
    switch (Context)
    {
    case DIC_GLOBAL:
        assert(CurGlobal != NULL);
        CurGlobal->classe = classe;
        break;
    case DIC_LOCAL:
        assert(CurLocal != NULL);
        CurLocal->classe = classe;
        break;
    }
}

//recupere l'information indique de l'entree courante
SYMType dicGetType()
{
    switch (Context)
    {
    case DIC_GLOBAL:
        assert(CurGlobal != NULL);
        return CurGlobal->type;
    case DIC_LOCAL:
        assert(CurLocal != NULL);
        return CurLocal->type;
    case DIC_ARGUMENTS:
        assert(CurArgument != NULL);
        return CurArgument->type;
    }
}
SYMClass dicGetClass()
{
    switch (Context)
    {
    case DIC_GLOBAL:
        assert(CurGlobal != NULL);
        return CurGlobal->classe;
    case DIC_LOCAL:
        assert(CurLocal != NULL);
        return CurLocal->classe;
    case DIC_ARGUMENTS:
        assert(CurArgument != NULL);
        return CurArgument->classe;
    }
}

int dicGetId()
{
    switch (Context)
    {
    case DIC_GLOBAL:
        assert(CurGlobal != NULL);
        return CurGlobal->id;
    case DIC_LOCAL:
        assert(CurLocal != NULL);
        return CurLocal->id;
    case DIC_ARGUMENTS:
        assert(CurArgument != NULL);
        return CurArgument->id;
    }
}

char *dicGetName()
{
    switch (Context)
    {
    case DIC_GLOBAL:
        assert(CurGlobal != NULL);
        return CurGlobal->name;
    case DIC_LOCAL:
        assert(CurLocal != NULL);
        return CurLocal->name;
    case DIC_ARGUMENTS:
        assert(CurArgument != NULL);
        return CurArgument->name;
    }
}

//ajoute une paire de bornes au tableau de l'entree courante
void dicSetArray(int inf,int sup)
{
    assert(Context != DIC_ARGUMENTS);
    switch (Context)
    {
    case DIC_GLOBAL:
        assert(CurGlobal != NULL);
        assert(CurGlobal->classe == CLASS_ARRAY);
        assert(CurGlobal->infos.bornes != NULL);
        CurGlobal->infos.bornes->BorneInf=inf;
        CurGlobal->infos.bornes->BorneSup=sup;
        break;

    case DIC_LOCAL:
        assert(CurLocal != NULL);
        assert(CurLocal->classe == CLASS_ARRAY);
        assert(CurLocal->infos.bornes != NULL);
        CurLocal->infos.bornes->BorneInf=inf;
        CurLocal->infos.bornes->BorneSup=sup;
        break;
    }
}

//force NULL a la liste des bornes
void dicVoidArray()
{
    assert(Context != DIC_ARGUMENTS);
    switch (Context)
    {
    case DIC_GLOBAL:
        assert(CurGlobal != NULL);
        assert(CurGlobal->classe == CLASS_ARRAY);
        CurGlobal->infos.bornes = NULL;
        break;

    case DIC_LOCAL:
        assert(CurLocal != NULL);
        assert(CurLocal->classe == CLASS_ARRAY);
        CurLocal->infos.bornes = NULL;
        break;
    }
}


//renvoie un pointeur sur la dim-ieme paire de bornes
//Bornes* dicGetArray(int dim=1)
Bornes* dicGetArray()
{
    switch (Context)
    {
    case DIC_GLOBAL:
        assert(CurGlobal != NULL);
        assert(CurGlobal->classe == CLASS_ARRAY);
        return CurGlobal->infos.bornes;
        break;

    case DIC_LOCAL:
        assert(CurLocal != NULL);
        assert(CurLocal->classe == CLASS_ARRAY);
        return CurLocal->infos.bornes;
        break;

    case DIC_ARGUMENTS:
        assert(CurArgument != NULL);
        assert(CurArgument->classe == CLASS_ARRAY);
        return CurArgument->infos.bornes;
        break;
    }
}


//definit si l'entree courant est un passage par valeur ou par reference 
//(par default: ARG_VALUE en global et ARG_REFERENCE en local)
void dicSetRef(ARGType r)
{
    assert(Context != DIC_ARGUMENTS);
    switch (Context)
    {
    case DIC_GLOBAL:
        assert(CurGlobal != NULL);
        assert(CurGlobal->classe == CLASS_VARIABLE);
        CurGlobal->infos.ref = r;
        break;

    case DIC_LOCAL:
        assert(CurLocal != NULL);
        assert(CurLocal->classe == CLASS_VARIABLE);
        CurLocal->infos.ref = r;
        break;
    }
}

//renvoie le type de passage de la variable
ARGType dicGetRef()
{
    switch (Context)
    {
    case DIC_GLOBAL:
        assert(CurGlobal != NULL);
        assert(CurGlobal->classe == CLASS_VARIABLE);
        return CurGlobal->infos.ref;
        break;

    case DIC_LOCAL:
        assert(CurLocal != NULL);
        assert(CurLocal->classe == CLASS_VARIABLE);
        return CurLocal->infos.ref;
        break;

    case DIC_ARGUMENTS:
        assert(CurArgument != NULL);
        assert(CurArgument->classe == CLASS_VARIABLE);
        return CurArgument->infos.ref;
        break;
    }
}

//incremente le nombre d'arguments de la procedure courante
void dicIncArgsQuantity()
{
    assert(Context != DIC_ARGUMENTS);
    assert(Context != DIC_LOCAL);

    assert(CurGlobal != NULL);
    assert(CurGlobal->classe == CLASS_PROCEDURE);

    CurGlobal->infos.nArguments += 1;
}

//recupere le nombre d'arguments
int dicGetArgsQuantity()
{
    assert(Context != DIC_ARGUMENTS);
    switch (Context)
    {
    case DIC_GLOBAL:
        assert(CurGlobal != NULL);
        assert(CurGlobal->classe == CLASS_PROCEDURE);
        return CurGlobal->infos.nArguments;
        break;

    case DIC_LOCAL:
        assert(CurLocal != NULL);
        assert(CurLocal->classe == CLASS_PROCEDURE);
        return CurLocal->infos.nArguments;
        break;
    }
}

//modifie ou recupere l'offset par rapport au sommet de pile
void dicSetOffset(int offset)
{
    assert(Context != DIC_ARGUMENTS);
    switch (Context)
    {
    case DIC_GLOBAL:
        assert(CurGlobal != NULL);
        CurGlobal->sp_offset=offset;
        break;

    case DIC_LOCAL:
        assert(CurLocal != NULL);
        CurLocal->sp_offset=offset;
        break;
    }
}
int dicGetOffset()
{
    switch (Context)
    {
    case DIC_GLOBAL:
        assert(CurGlobal != NULL);
        return CurGlobal->sp_offset;
    case DIC_LOCAL:
        assert(CurLocal != NULL);
        return CurLocal->sp_offset;
    case DIC_ARGUMENTS:
        assert(CurArgument != NULL);
        return CurArgument->sp_offset;
    }
}

//ajoute une nouvelle entree au dictionnaire de type et de classe definis dans le constructeur
//et la selectionne
void dicAddEntry(char *name)
{

    DicElem *newelem = malloc(sizeof(DicElem));
    char *newName =  malloc(sizeof(strlen(name) +1 ));

    //preparation de la nouvelle entree
    strcpy(newName,name);
    newelem->name=newName;

    idCount = idCount + 1;
    newelem->id=idCount ;
    newelem->sp_offset=BuilderOffset;
    newelem->type=BuilderType;
    newelem->classe=BuilderClass;
    newelem->procArgList=NULL;

    switch(BuilderClass)
    {
        case CLASS_VARIABLE:
            newelem->infos.ref=BuilderRef;   
            break;

        case CLASS_PROCEDURE:
            newelem->infos.nArguments=0; //increment plus tard
            break;
            
        case CLASS_ARRAY:
            newelem->infos.bornes= malloc(sizeof(Bornes)); 
            break; //devra etre modifie pour les multi dimensions
    }
    newelem->next=NULL; //ajoute en fin de liste : pas de suivant

    //insertion dans le dictionnaire
    switch(Context)
    {
        case DIC_GLOBAL:
            CurGlobal=Global;
            if(CurGlobal != NULL)
            {
                while(CurGlobal->next != NULL)
                {
                    CurGlobal = CurGlobal->next;
                }
                CurGlobal->next = newelem;
                CurGlobal = CurGlobal->next;
            }
            else
            {
                CurGlobal = newelem;
                Global = newelem;
            }            
            break;

        case DIC_LOCAL:
            CurLocal=Local;
            if(CurLocal != NULL)
            {
                while(CurLocal->next != NULL)
                {
                    CurLocal = CurLocal->next;
                }
                CurLocal->next = newelem;
                CurLocal = CurLocal->next;
            }
            else
            {
                CurLocal = newelem;
                Local = newelem;
            } 
            break;
    }
    
}

static void wipeArgList(DicElem *courant)
{
	assert(courant != NULL);
	DicElem * suivant=courant->next;

	while(courant != NULL)
	{
		assert(courant->classe != CLASS_PROCEDURE);
		killContent(courant);
		free(courant);
		
		courant=suivant;
		if(suivant != NULL)
			suivant=suivant->next;
	}
}
//nettoie les elements dynamiques de l'element
static void killContent(DicElem *courant)
{
    assert(courant != NULL);
    assert(courant->name != NULL);
    free(courant->name);

    if(courant->classe == CLASS_ARRAY)
    {
        assert(courant->infos.bornes != NULL);
        free(courant->infos.bornes);
		courant->infos.bornes=NULL;
    }
	
	if(courant->classe == CLASS_PROCEDURE)
    {
        if(courant->procArgList != NULL)
        	wipeArgList(courant->procArgList);
    }
}

//efface le dictionnaire local 
void dicCleanLocal()
{
    DicElem * coupe=NULL;
    DicElem * courant=NULL;
    DicElem * suivant=NULL;
    int nArgs = CurGlobal->infos.nArguments;

    //deja vide?
    if(Local==NULL)
    {
        return;
    }

    CurLocal=Local;
    //sauvegarde des arguments
    if(nArgs > 0)
    {
        while(nArgs > 0)
        {
            assert(CurLocal != NULL);
            coupe = CurLocal;
            CurLocal = CurLocal->next;
            nArgs = nArgs-1;
        }

        assert(CurGlobal->classe == CLASS_PROCEDURE);
        CurGlobal->procArgList = Local;
        coupe->next = NULL;
    }
    else
    {
        assert(CurGlobal->classe == CLASS_PROCEDURE);
        CurGlobal->procArgList = NULL;
    }

    //nettoyage des locales
    if(CurLocal != NULL)
    {
        courant = CurLocal;
        suivant = courant->next;
    }
    while(courant != NULL)
    {
        killContent(courant);
        free(courant);
        courant = suivant;
		if(suivant!=NULL)
        	suivant = suivant->next;
    }
    CurLocal = NULL;
    Local = NULL;
}
//recherche l'existence du nom, renvoie 1 si il existe dans le contexte courant, 0 sinon
int dicCheckExisting(char *nom)
{
    DicElem *eTmp;
    int result = 0;

    switch(Context)
    {
        case DIC_GLOBAL:
            eTmp=Global;
            break;
        case DIC_LOCAL:
            eTmp=Local;
            break;
        case DIC_ARGUMENTS:
            eTmp=Argument;
            break;
    }

    if(eTmp != NULL)
    {
        while(eTmp->next != NULL && result == 0)
        {
            if(strcmp(eTmp->name,nom) == 0)
                result = 1;
            if(result == 0)
            {
                eTmp = eTmp->next;
            }
        }
        //test du dernier element
        if(result == 0)
        {
            if(strcmp(eTmp->name,nom) == 0)
                result = 1;
        }

    }
    return result;           
}

//rend courante l'entree du nom recherche dans le contexte courant, renvoie 0 si non trouve
int dicSelect(char *nom)
{
    DicElem *eTmp;
    int result = 0;

    switch(Context)
    {
        case DIC_GLOBAL:
            eTmp=Global;
            break;
        case DIC_LOCAL:
            eTmp=Local;
            break;
        case DIC_ARGUMENTS:
            eTmp=Argument;
            break;
    }

    if(eTmp != NULL)
    {
        while(eTmp->next != NULL && result == 0)
        {
            if(strcmp(eTmp->name,nom) == 0)
                result = 1;
            if(result == 0)
            {
                eTmp = eTmp->next;
            }
        }
        //test du dernier element
        if(result == 0)
        {
            if(strcmp(eTmp->name,nom) == 0)
                result = 1;
        }
    }
    if(result == 1)
    {
        switch(Context)
        {
        case DIC_GLOBAL:
            CurGlobal = eTmp;
            break;
        case DIC_LOCAL:
            CurLocal = eTmp;
            break;
        case DIC_ARGUMENTS:
            CurArgument= eTmp;
            break;
        }
    }
    return result; 
}

int dicSelectFirst()
{
    switch(Context)
    {
        case DIC_GLOBAL:
            if(Global != NULL)
                CurGlobal = Global;
            else
                return 0;
            break;
        case DIC_LOCAL:
            if(Local != NULL)
                CurLocal = Local;
            else
                return 0;
            break;
        case DIC_ARGUMENTS:
            if(Argument != NULL)
                CurArgument = Argument;
            else
                return 0;
            break;
        }
    return 1;
}

int dicSelectNext()
{
    switch(Context)
    {
        case DIC_GLOBAL:
            if(CurGlobal != NULL )
            {
                if(CurGlobal->next != NULL)
                    CurGlobal = CurGlobal->next;
                else
                    return 0;
            }   
            else
                return 0;
            break;

        case DIC_LOCAL:
            if(CurLocal != NULL )
            {
                if(CurLocal->next != NULL)
                    CurLocal = CurLocal->next;
                else
                    return 0;
            }   
            else
                return 0;
            break;

        case DIC_ARGUMENTS:
            if(CurArgument != NULL )
            {
                if(CurArgument->next != NULL)
                    CurArgument = CurArgument->next;
                else
                    return 0;
            }   
            else

                return 0;
            break;
        }
    return 1;
}
//charge la liste des arguments de la proc selectionnee
void dicLoadArgList()
{
    assert(Context == DIC_GLOBAL);
    assert(CurGlobal->classe == CLASS_PROCEDURE);

    Argument = CurGlobal->procArgList;
    CurArgument = Argument;
}

//grand nettoyage de printemps, on vire tout partout
void dicMrProper()
{
    DicElem * courant=NULL;
    DicElem * suivant=NULL;

	//nettoyage du global
	if(Global != NULL)
	{
		courant=Global;
		suivant=courant->next;

		while(courant != NULL)
		{
			
			killContent(courant);
			
			assert(courant != NULL);
			free(courant);
			
			courant=suivant;
			if(suivant != NULL)
				suivant=suivant->next;
		}
		
		Global=NULL;
		CurGlobal=NULL;
	}

	//nettoyage du Local
	if(Local != NULL)
	{
		courant=Local;
		suivant=courant->next;
		
		while(courant != NULL)
		{
			
			killContent(courant);
			
			assert(courant != NULL);
			free(courant);
			
			courant=suivant;
			if(suivant != NULL)
				suivant=suivant->next;
		}
		Local=NULL;
		CurLocal=NULL;
	}
	
	//rien a nettoyer pour les arguments: reference vers un point du global
	CurArgument=NULL;
	Argument=NULL;
}
//renvoie 0 si n'existe pas, le numero du contexte
//ou il a ete trouve sinon
//ne change pas le contexte courant
DICContext dicSelectReachable (char *nom)
{
  DICContext OldContext = dicGetContext ();
  DICContext retour = 0;

  assert(OldContext != DIC_ARGUMENTS);
  
  if (dicSelect (nom))
  {
    retour = OldContext;
  }

  if (OldContext == DIC_LOCAL && retour == 0 )
  {
    dicSetContext (DIC_GLOBAL);
    if (dicSelect (nom))
    {
      retour = DIC_GLOBAL;
    }
    dicSetContext (OldContext);
  }

  return retour;
}

//renvoie 0 si n'existe pas dans tout le dictionnaire
//1 sinon
//ne change pas le contexte courant
int dicCheckExistingAtAll (char *nom)
{
  DICContext OldContext = dicGetContext ();
  int retour = 0;

  assert(OldContext != DIC_ARGUMENTS);
  retour = dicCheckExisting (nom);

  if (OldContext == DIC_LOCAL  && retour == 0 )
  {
    dicSetContext (DIC_GLOBAL);
    retour = dicCheckExisting (nom);
    dicSetContext (OldContext);
  }
  return retour;
}

/*selectionne le nArg'ieme argument de 'procedure'
  et reste dans le contexte DIC_ARGUMENTS
*/
void dicSelectArgument (int nArg, char *procedure)
{
    int i;

    dicSetContext(DIC_GLOBAL);
    
    assert(procedure);
    dicSelect(procedure);
    assert(nArg <= dicGetArgsQuantity());
    assert(nArg > 0);

    dicLoadArgList ();
    dicSetContext(DIC_ARGUMENTS);


    dicSelectFirst ();
    while(nArg > 1)
    {
        dicSelectNext();
        nArg = nArg - 1;
    }
}
