#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include "unitesLexicales.h"
#include <string.h>
#include "dictionnaire.h"
#include "gen.h"
#include "piletypes.h"

/*************Indentation*********************/
/* indent <input> -o <output> -bc -bl0 -bli0 -brs -cbi0 -cli2 -i2 -l80 -lc80 -npsl -sc -ss */


/*Variables globales*/
int uc;

/*Variables globales importï¿œs de l'analyseur lexical*/
extern int noligne;
extern int valeur;
extern float valeurf;
extern char *yytext;
extern FILE *yyin;

/*Prototypes des fonctions de l'analyseur syntaxique*/
void math_expr (void);
void id_list (void);
void liste_bornes (void);
void id_tableau (void);
void for_list (void);
void expr_list (void);
void list_arg_ext (void);
void arg_list (int *cptArgs);
void id_suivant (void);
void id_simple (void);
void proc (void);
void id_tableau (void);
void declaration (void);

void body (void);
void block (void);
void program (void);

/*Variables globales memorisation des type et classe*/
SYMType memType;
SYMClass memClass;
ARGType memRef;

//Memorisation de l'identifieur de partie gauche (section instructions)
char * instrIdent;
SYMType instrType;
SYMClass instrClass;
DICContext instrContext;


/*Fonctions Test dico*/
void
printDatas ()
{
	printf ("%s->id:%d , type:%d, classe :%d, sp_offset:%d",
		dicGetName (), dicGetId (), dicGetType (), dicGetClass (),
		dicGetOffset ());
	switch (dicGetClass ())
	{
	case CLASS_VARIABLE:
		printf (", ref:%d", dicGetRef ());
		break;
	case CLASS_PROCEDURE:
		printf (", nArgs:%d", dicGetArgsQuantity ());
		break;
	case CLASS_ARRAY:
    if(dicGetArray() != NULL)
    {
      printf (", borneinf:%d, bornesup:%d",
			dicGetArray ()->BorneInf, dicGetArray ()->BorneSup);
    }
		break;
	}
	printf ("\n");
}

void
printLocalDic ()
{
	dicSetContext (DIC_GLOBAL);
	printf ("Dictionnaire du bloc %s :\n", dicGetName ());
	dicSetContext (DIC_LOCAL);

	if (dicSelectFirst ())
	{
		printDatas ();
		while (dicSelectNext ())
			printDatas ();
	}
}

void
printGlobalDic ()
{
	dicSetContext (DIC_GLOBAL);
	printf ("Dictionnaire global:\n");

	if (dicSelectFirst () != 0)
		printDatas ();

	while (dicSelectNext ())
		printDatas ();
}


/*Fin test dico*/

/*Fonctions utilitaires*/
/*Fonction erreur:
Affiche un message sur la sortie standard d'erreur et arrete le processus
Necessite la variable importee noligne en global*/
void
erreur (char *message)
{
	fprintf (stderr, "ERROR in line %d: %s\n", noligne, message);
	exit (EXIT_FAILURE);
}

void warning(char* message)
{
  	fprintf (stderr, "WARNING in line %d: %s\n", noligne, message);
}

/*Fonctions de l'analyseur syntaxique*/

/*Section expressions mathematiques*/

//Cette fonction produit le code necessaire pour caster le type le plus general
//a deux operandes (a appeler avant chaque operation a deux operandes)

static SYMType
equalizeTypes (void)
{
	SYMType type = SYM_INTEGER;
	assert (!stypeIsEmpty ());
	if (stypeGet () == SYM_REAL)
	{
	//Si le 2eme est un int on le caste en float
		type = SYM_REAL;
		if (stypeGetSecond () == SYM_INTEGER)
		{
			//Caster le 2eme element au sommet de la pile en float
			IntToFloatSecond ();
		}  
	}
	else
	{
		if (stypeGet () == SYM_INTEGER)
		{
			//Si le 2eme est un float, on caste le 1er en float
			if (stypeGetSecond () == SYM_REAL)
			{
				//Caster le 1er element de la pile en float
				IntToFloat ();
				type = SYM_REAL;
			}
		}
	}
	return type;
}

/*verifie que le cptArgs'ieme argument de la procedure 'identificateur'
corresponde a ce qui est en cours d'analyse dans le yylex.
identificateur: procedure appelee
CurrentContext: contexte de l'appel a la procedure
argContext: contexte de l'argument en cours d'analyse
*/
static void
argControl (int cptArgs, char *identificateur, DICContext CurrentContext)
{
  DICContext argContext=0;
  
	dicSelectArgument(cptArgs, identificateur);
 
	//rejets semantiques en cas de passage par reference
	//seuls acceptes: identificateur de tableau ou de variable simple
  if(dicGetClass() == CLASS_ARRAY || (dicGetClass() == CLASS_VARIABLE 
                                      && dicGetRef() == ARG_REFERENCE))
	{  
    dicSetContext (CurrentContext); 
    
		if (uc != IDENT)
			erreur ("Identifier requested for reference-passing argument");
		argContext = dicSelectReachable (yytext) ;
    
		if (argContext == 0)
			erreur ("Unknown identifier (argControl)");
    
		if (dicGetClass () == CLASS_PROCEDURE)
			erreur ("Unable to use a procedure result as a reference");
    
		//Empiler l'adresse vers la variable identificateur
		dicSetContext (argContext);
    if(dicGetClass() == CLASS_ARRAY)
      SetTabHeadStack(yytext);
    else
      SetReferenceStack (yytext);
		stypeStack (dicGetType());
    
		dicSetContext (CurrentContext);    
		uc = yylex ();
		if(uc != COMMA && uc != BRACKETR)
			erreur("unauthorized operation for reference-passing argument");
	}
	else
	{          
    dicSetContext (CurrentContext); 
        
		//On peut calculer
		math_expr ();
	}
	//Le sommet de pile contient la valeur de l'argument
	//comme la pile virtuelle contient le type
	//Controler le type avec l'argument de la procedure
	dicSelectArgument (cptArgs,identificateur);
	if (dicGetType () != stypeGet ())
		erreur ("effective argument type does not match required type");
	dicSetContext (CurrentContext);
}


void
facteur (void)
{
	char *identificateur;
	DICContext CurrentContext = dicGetContext ();
	DICContext identContext=0;
	SYMType type;
	ARGType argtype;
	int cptArgs = 0;
  
  int negate =0;
  
  if(uc == MINUS)
  {
    negate = 1;
    uc = yylex (); // 4 cas: var int, var float, immediat, autre
  }
  
	switch (uc)
	{
	case BRACKETL:
		uc = yylex ();
		math_expr ();

		if (uc != BRACKETR)
			erreur ("missing ) after end of bracketted expression");

		uc = yylex ();
		break;
    
	case NBR_INT:
		//On lit un entier en dur : on l'empile et on renseigne le type
		SetConstIntStack (valeur);
		stypeStack (SYM_INTEGER);
		uc = yylex ();
		break;

	case NBR_REAL:
		//On lit un réel en dur : on l'empile et on renseigne le type
		SetConstFloatStack (valeurf);
		stypeStack (SYM_REAL);
		uc = yylex ();
		break;
    
  case BTRUE:
    SetConstBoolStack (1);
		stypeStack (SYM_BOOLEAN);
    uc = yylex ();
		break;
                
  case BFALSE:
    SetConstBoolStack (0);
		stypeStack (SYM_BOOLEAN);
    uc = yylex ();
		break;
  
	case IDENT:

		//Sauvegarder l'identifieur invoqué dans la variable locale identificateur
		//identificateur = strcpy (identificateur, yytext);
		//Trouver identificateur
		identContext = dicSelectReachable (yytext);
		if (!identContext)
			erreur ("unknown identifier (facteur)");
		dicSetContext(identContext);
		identificateur = dicGetName();
		//Empiler le type de l'identificateur (ce sera la valeur qui residera
		//au sommet de pile apres les opérations necessaire pour empiler sa valeur
		stypeStack (dicGetType ());
		dicSetContext (CurrentContext);

		uc = yylex ();
		if (uc == ARRAYBRACKETL)
		{
			//Gestion des tableaux
			uc = yylex ();
			math_expr ();

			if(stypeGet() != SYM_INTEGER)
				erreur("Indexation needs INTEGER value");
			
			GetArrayValueStack (identificateur);
			
			stypeUnstack();
			
			if (uc != ARRAYBRACKETR)
				erreur ("missing ] at the end of array indexation expression");
			uc = yylex ();
		}
		else
		{
			if (uc == BRACKETL)
			{	//Appel de fonction

				if (dicGetClass () != CLASS_PROCEDURE)
					erreur ("Unable to call a non-procedure class data");
        
				//Verifier que le type de la procedure n'est pas void
				if (dicGetType () == SYM_VOID)
					erreur ("Unable to call a void procedure in an expression");
        
				//Prendre le sommet de la pile pour calculer les expressions
				//Controler les arguments (le bon type, la bonne classe, le bon nombre)
				CallProc(identificateur);
				uc = yylex ();
				if (uc != BRACKETR)
				{
					//gestion du premier argument

					cptArgs++;
					//Controler l'argument effectif
					argControl (cptArgs,identificateur,CurrentContext);
					//arguments suivants
					while (uc == COMMA)
					{
						//Gestion des arguments suivants
						uc = yylex ();

						cptArgs++;
						argControl (cptArgs,identificateur,CurrentContext);
					}

					//verification du compte des arguments
					dicSelect (identificateur);
					if (dicGetArgsQuantity () != cptArgs)
						erreur ("Argument count mismatch");

					if (uc != BRACKETR)
						erreur ("missing ) at the end of function parameter in expression");

					uc = yylex ();
				}
				else
				{
					//verifier que nbArgs de la procedure identificateur = 0
					if(dicGetArgsQuantity () != 0)
						erreur ("Some arguments are missing");
					//Si il y a BRACKETR on avance (fonction sans parametres)
					uc = yylex ();
				}
				EndCallProc(identificateur);
			}

			else
			{
				//Identificateur variable simple
				SetStack (identificateur);
			}

		}
		break;

	default:
		erreur ("unauthorized operation attempted in math expression, aborting evaluation");
		break;
	}
    
  if(negate)
  {
    switch(stypeGet())
    {
      case SYM_INTEGER:
        NegInt();
        break;
      case SYM_REAL:
        NegFloat();
        break;
      default:
        erreur("cant negate something else than an integer or a real value");
    }
  }
}

void
terme (void)
{
	//oper : Sauvegarde du signe de l'operation
	int oper;
	SYMType type;
 
	facteur ();
  
	while (uc == MULT || uc == DIV || uc == DIVENT)
	{

		//oper : Sauvegarde du signe d'operation
		oper = uc;
		uc = yylex ();

		facteur ();

		//Changer les type,
		//garder int si possible, si l'un des deux est un float, caster l'int en float
		//Le cast va donner le type de l'opération (opération d'entiers ou de reels
		assert (!stypeIsEmpty ());
		type = equalizeTypes();
		//Faire l'operation definie selon le signe
		switch (oper)
		{
		case MULT:
			stypeUnstack ();
			stypeUnstack ();
			if (type == SYM_REAL)
			{
				MulFloat ();
				stypeStack (type);
			}
			if (type == SYM_INTEGER)
			{
				MulInt ();
				stypeStack (type);
			}
			break;
		case DIV:
			stypeUnstack ();
			stypeUnstack ();
			if (type == SYM_REAL)
			{
				DivFloat ();
				stypeStack (SYM_REAL);
			}
			if (type == SYM_INTEGER)
			{
				//caster les 2 entiers du haut de la pile en reel (division reelle)
				IntToFloat ();
				IntToFloatSecond ();
				DivFloat ();
				stypeStack (SYM_REAL);
			}
			break;
		case DIVENT:
			stypeUnstack ();
			stypeUnstack ();
			if (type == SYM_REAL)
			{
				//caster les 2 floats du haut de la pile en int (division entiere)
				FloatToInt ();
				FloatToIntSecond ();
				DivInt ();
				stypeStack (SYM_INTEGER);
			}
			if (type == SYM_INTEGER)
			{
				DivInt ();
				stypeStack (SYM_INTEGER);
			}
			break;
		}
	}
}

void
expArith (void)
{
	//oper : Sauvegarde du signe de l'operation
	int oper;
	SYMType type = SYM_INTEGER;

	terme ();

	while (uc == PLUS || uc == MINUS)
	{
		//Sauvegarde du signe d'operation
		oper = uc;
		uc = yylex ();
		terme ();
		//Changer les type, garder int si possible, si l'un des deux est un float, caster l'int en float
		type = equalizeTypes();
		
		//Exécuter l'opération memorisee dans oper
		switch (oper)
		{
		case PLUS:
			stypeUnstack ();
			stypeUnstack ();
			if (type == SYM_REAL)
			{
				AddFloat ();
				stypeStack (type);
			}
			if (type == SYM_INTEGER)
			{
				AddInt ();
				stypeStack (type);
			}
			break;
		case MINUS:
			stypeUnstack ();
			stypeUnstack ();
			if (type == SYM_REAL)
			{
				SubFloat ();
				stypeStack (SYM_REAL);
			}
			if (type == SYM_INTEGER)
			{
				SubInt ();
				stypeStack (SYM_INTEGER);
			}
			break;
		}
	}
}

void
comparaison (void)
{
	int oper;
	SYMType type;
  
	expArith ();
  
	switch (uc)
	{
	case EQUAL:
	case NOTEQUAL:
	case LESS:
	case NOTGREATER:
	case NOTLESS:
	case GREATER:
		//Une comparaison fait que le type en sommet de pile devient un booleen
		//on sauvegarde l'operande de la comparaison
		oper=uc;
		uc = yylex ();
		expArith ();
		//Il faut faire correspondre les types
		type = equalizeTypes ();
		stypeUnstack ();
		stypeUnstack ();
		switch(oper)
		{
		case EQUAL:
			Equal();
			break;
		case NOTEQUAL:
			NotEqual();
			break;
		case LESS:
			LessThan();
			break;
		case NOTGREATER:
			LessThanEqual();
			break;
		case NOTLESS:
			GreaterThanEqual();
			break;
		case GREATER:
			GreaterThan();
			break;
		}
		stypeStack(SYM_BOOLEAN);
	default:
		;
	}
}

void
conjonction (void)
{
	comparaison ();
	while (uc == AND)
	{
		uc = yylex ();
		comparaison ();
		//verifier que les deux operandes sont de type de booleen, sinon erreur
		if(!(stypeGet () == SYM_BOOLEAN && stypeGetSecond () == SYM_BOOLEAN))
		{
			erreur("'AND' is a boolean-type logical operation, operand type mismatch");
		}
		//declencher la production de code de l'operation
		stypeUnstack ();
		And();
	}
}

void
math_expr (void)
{
	int notswitch = 0;
	if (uc == NOT)
	{
		uc = yylex ();
		notswitch = 1;
	}
	conjonction ();
	while (uc == OR)
	{
		uc = yylex ();
		conjonction ();
		//verifier que les deux operandes sont de type de booleen, sinon erreur
		if(!(stypeGet () == SYM_BOOLEAN && stypeGetSecond () == SYM_BOOLEAN))
		{
			erreur("'OR' is a boolean-type logical operation, operand type mismatch");
		}
		//declencher la production de code de l'operation
		stypeUnstack ();
		Or();
	}
	if(notswitch)
	{
		if(stypeGet () != SYM_BOOLEAN)
			erreur("'NOT' is a boolean-type logical operation, operand type mismatch");
		//stypeUnstack ();
		Not();
	}
}

/*Section Instructions*/
/*
void
for_type (void)
{
	if (uc == STEP)
	{
		uc = yylex ();
		math_expr ();
		if(stypeGet() != SYM_INTEGER)
			erreur ("'STEP' data must be an integer");
		if (uc == UNTIL)
		{
			uc = yylex ();
			math_expr ();
			if(stypeGet() != SYM_INTEGER)
				erreur ("'UNTIL' data must be an integer");
		}
		else
			erreur ("missing 'UNTIL' keyword with 'STEP'-type 'FOR' statement");
	}
	else
	{
		if (uc == WHILE)
		{
			//Pas encore gere
			erreur("'WHILE' Statement cannot be intrepreted yet");
			uc = yylex ();
			math_expr ();
		}
	}
}

void
for_list_element (void)
{
	math_expr ();
	if(stypeGet() != SYM_INTEGER)
		erreur("The value must be an integer");
	GetStack (instrIdent);
	VarFor();
	for_type ();
}

void
for_listc (void)
{
	switch (uc)
	{
	case COMMA:
		uc = yylex ();
		for_list ();
		break;
	case DO:
//LEQUEL EST LE BON entre ForDo et WhileDo
		ForDo();
		//WhileDo();
		
		uc = yylex ();
		break;
	default:
		erreur ("abnormal termination for list in 'FOR' statement");
	}
}

*/

void
for_list (void)
{
//	for_list_element ();
//	for_listc (); //methode recursive


//ex for_list_element
	math_expr ();
	if(stypeGet() != SYM_INTEGER)
		erreur("The value must be an integer");

	GetStack (instrIdent);
	VarFor();

//ex for_type
	if (uc == STEP)
	{
		uc = yylex ();
		math_expr ();
		if(stypeGet() != SYM_INTEGER)
			erreur ("'STEP' data must be an integer");
		if (uc == UNTIL)
		{
			uc = yylex ();
			math_expr ();
			if(stypeGet() != SYM_INTEGER)
				erreur ("'UNTIL' data must be an integer");
		}
		else
			erreur ("missing 'UNTIL' keyword with 'STEP'-type 'FOR' statement");

      if(uc != DO)
         erreur("DO delimiter is missing");

		ForDo(instrIdent);
	}
	else if (uc == WHILE)
  {
    //Pas encore gere
    //erreur("'WHILE' Statement cannot be intrepreted yet");
    While();
    uc = yylex ();
    math_expr ();
    
    if(uc != DO)
       erreur("DO delimiter is missing");
    
    WhileDo();
  }
	
//reste de feu for_listc
	uc = yylex ();
}

void
for_clause (void)
{
	DICContext CurrentContext;
	if (uc != IDENT)
		erreur ("missing identifier after 'FOR' statement");

	instrContext = dicSelectReachable(yytext);
	CurrentContext = dicGetContext();
	dicSetContext(instrContext);
	instrIdent = dicGetName();
	instrType = dicGetType ();
	instrClass = dicGetClass ();
	dicSetContext (CurrentContext);

	if(instrType != SYM_INTEGER )
		erreur ("'FOR' variable must be an integer");
	uc = yylex ();

	if (uc != AFFECT)
	{
		erreur ("missing := for identifier in 'FOR' statement");
	}
	uc = yylex ();
	for_list ();
}

void
else_bloc (void)
{
	if (uc == ELSE)
	{
		Else();
		uc = yylex ();
		body ();
		EndElse();
	}
}

void
then_bloc (void)
{
	if (uc == THEN)
	{
		Then();
		uc = yylex ();
		body ();
		EndThen();
	}
	else
		erreur ("missing then before statement block");
}

void
expr_suiv ()
{
	if (uc == COMMA)
	{
		uc = yylex ();
		expr_list ();
	}
	else
	{
		if (uc == BRACKETR)
			uc = yylex ();
		else
			erreur ("missing , or ) in expression list");
	}
}

void
expr_list (void)
{
	if (uc == OPENSTRING)
	{
		uc = yylex ();
	}
	else
		math_expr ();
	expr_suiv ();
}


void
arg_suivant (int *cptArgs)
{
  DICContext callContext = dicGetContext();
  
	if (uc == COMMA)
	{
		uc = yylex ();
		arg_list (cptArgs);
	}
	else
	{
		if (uc == BRACKETR)
		{
          
      //verification du compte des arguments
      dicSetContext(DIC_GLOBAL);
      dicSelect (instrIdent);
      if (dicGetArgsQuantity () != *cptArgs)
        erreur ("Argument count mismatch");
      dicSetContext(callContext);
      
			uc = yylex ();
			if (uc != COMMADOT)
				erreur ("Missing ; after argument list");
			uc = yylex ();
		}
	}
}

void
arg_list (int *cptArgs)
{
  DICContext callContext = dicGetContext();
        
	if (uc == BRACKETR)
	{
		//erreur, appel du type proc(A,)
	    if (*cptArgs != 0)
			erreur ("Argument list error in procedure");
		//verification du compte des arguments
		dicSetContext(DIC_GLOBAL);
		dicSelect (instrIdent);
		if (dicGetArgsQuantity () != 0)
			erreur ("Where are my arguments?");
		dicSetContext(callContext);
		EndCallProc(instrIdent);
		uc = yylex ();
		if (uc != COMMADOT)
			erreur ("Missing ; after argument list");
		uc = yylex ();
	}
	else
	{
	    *cptArgs = *cptArgs +1;
    
		argControl (*cptArgs, instrIdent, dicGetContext());
          
		arg_suivant (cptArgs);
	}
}

static void implicitCasts()
{
  switch(instrType)
  {
    case SYM_INTEGER:
      if(stypeGet() == SYM_REAL)
      {
        FloatToInt();
      }
      else if (stypeGet() == SYM_BOOLEAN)
      {
        warning("implicit cast INTEGER -> BOOLEAN");
      }
      else
      {
        erreur("non-sense cast");
      }
      stypeUnstack();
      stypeStack(SYM_INTEGER);
      break;
    
    case SYM_BOOLEAN:
      if(stypeGet() == SYM_REAL)
      {
        warning("implicit cast REAL -> BOOLEAN, can produce fake FALSE");
        FloatToInt();
      }
      else if (stypeGet() == SYM_INTEGER)
      {
      }
      else
      {
        erreur("non-sense cast");
      }
      stypeUnstack();
      stypeStack(SYM_REAL);
      break;
    
    case SYM_REAL:
      if(stypeGet() == SYM_BOOLEAN)
      {
        warning("implicit cast BOOLEAN -> REAL");
        IntToFloat();
      }
      else if (stypeGet() == SYM_INTEGER)
      {
        IntToFloat();
      }
      else
      {
        erreur("non-sense cast");
      }
      stypeUnstack();
      stypeStack(SYM_BOOLEAN);
      break;
  }
}

void
voidproc_affect (void)
{
  int cptArgs=0;
	if (uc == AFFECT)
	{
		uc = yylex ();
		math_expr ();
    
		//verifier si le type correspond au type de l'identificateur membre gauche
      //if(instrType != stypeGet())
      //	erreur ("Unauthorized operation : implicit cast");
    if(instrType != stypeGet())
      implicitCasts();
    
		//realiser l'affectation
		switch(instrClass)
		{
		case CLASS_ARRAY:
			SetArrayValueStack(instrIdent);
			stypeUnstack();
			stypeUnstack();
			break;
		case CLASS_VARIABLE:
			GetStack(instrIdent);
			stypeUnstack();
			break;
      
    case CLASS_PROCEDURE: //affectation de retours de procedure
      //marche uniquement si la procedure a une valeur de retour: tester types
      assert(dicGetContext() == DIC_LOCAL); //ces affectations ne sont valides qu'en local
         
      if(instrType != stypeGet())
        erreur("implicit cast in procedure return affectation");
      
			GetStack(instrIdent);
			stypeUnstack();
			break;
      
		}
		
		if (uc == COMMADOT)
			uc = yylex ();
		else
			erreur ("missing ; at the end of affectation statement");
	}
	else
	{
		if (uc == BRACKETL)
		{
			CallProc(instrIdent);
			stypeStack(instrType);
			uc = yylex ();
			arg_list (&cptArgs);
		}
		else
			erreur ("Missing := or ( in identifier-beginning statement");
	}
}

void
tab (void)
{
	if (uc == ARRAYBRACKETL)
	{
		uc = yylex ();
		math_expr ();
		if(stypeGet() != SYM_INTEGER)
			erreur("An array can be indexed only by a integer value");
		if(instrClass != CLASS_ARRAY)
			erreur("This identifier leads to a non-array data");
		if (uc != ARRAYBRACKETR)
			erreur ("missing ] at the end of table indexation expression");
		uc = yylex ();
	}
}

void
instr (void)
{
	DICContext CurrentContext=0;
	DICContext identContext=0;
//	char* forVarName=NULL;
	
	CurrentContext=dicGetContext();
	//Liste des instructions, for, if et identifiants
	switch (uc)
	{
	case FOR:
		For();
		uc = yylex ();

		//Verification semantique: existence
		identContext = dicSelectReachable(yytext);
		if(!identContext)
			erreur("unknown identifier (instr)");
		dicSetContext(identContext);
		instrIdent = dicGetName();
//		forVarName = dicGetName();
		instrType = dicGetType();
		instrClass = dicGetClass();
		dicSetContext(CurrentContext);


		for_clause ();
		body ();
		EndFor();
		break;
	case IF:
		If();
		uc = yylex ();
		math_expr ();
		then_bloc ();
		else_bloc ();
		EndIf();
		break;
	case IDENT:
		//Verification semantique : procedures reservees
    //Gestion des procedures reservees
    if(strcmp(yytext, "WRITE") == 0) //routine WRITE predefinie
    {
      uc = yylex();
      if(uc != BRACKETL)
        erreur("missing ( after procedure WRITE call");
      uc = yylex();
      if(uc != OPENSTRING)
        erreur("WRITE procedure needs a between-\" string as argument");
      DisplayString(yytext);
      uc = yylex();
      if(uc != BRACKETR)
        erreur("WRITE procedure usage : WRITE(OPENSTRING);");
      uc=yylex();
      if(uc != COMMADOT)
        erreur("missing ; after WRITE statement");
      uc=yylex();
    }
    
    else if(strcmp(yytext, "PRINT") == 0) //routine PRINT predefinie
    {
      uc = yylex();
      if(uc != BRACKETL)
        erreur("missing ( after procedure PRINT call");
      uc = yylex();
/*
      math_expr();
      DisplayStackHead(stypeGet());
*/
      
      switch(uc)
      {
        case NBR_INT :
          DisplayInt(valeur);
          break;
        
        case NBR_REAL :
          DisplayFloat(valeurf);
          break;

        case IDENT :
          identContext = dicSelectReachable(yytext);
          if(identContext == 0)
            erreur("unknown identifier in PRINT call");  
          
          dicSetContext(identContext);
          switch(dicGetType())
          {
            case SYM_INTEGER:
              DisplayVarInt(yytext);
              break;
            case SYM_REAL:
              DisplayVarFloat(yytext);
              break;
            case SYM_BOOLEAN:
              DisplayVarBool(yytext);
              break;
          }
          break;
        default:
          erreur("PRINT procedure needs a variable name or a constant value as argument");

      }

      uc = yylex();
      
      
      if(uc != BRACKETR)
        erreur("PRINT procedure usage : PRINT(VAR_IDENT);");
      uc=yylex();
      
      if(uc != COMMADOT)
        erreur("missing ; after PRINT statement");
      uc=yylex();
    }
    
    else if(strcmp(yytext, "READ") == 0) //routine READ predefinie
    {
      uc = yylex();
      if(uc != BRACKETL)
        erreur("missing ( after procedure READ call");
      uc = yylex();
      if(uc != IDENT)
        erreur("READ procedure needs a variable name as argument");
      //mettre l'entrée dans la variable
      
      /** A COMPLETER **/
      erreur("READ not supported yet");
      
      uc = yylex();
      if(uc != BRACKETR)
        erreur("READ procedure usage : READ(VAR_IDENT);");
      uc=yylex();
      if(uc != COMMADOT)
        erreur("missing ; after READ statement");
      uc=yylex();
    }
   
		else //identifiant de variable
		{
			//Verification semantique: existence
			identContext = dicSelectReachable(yytext);
			if(!identContext)
				erreur("unknown identifier (instr)");
			dicSetContext(identContext);
			instrIdent = dicGetName();
			instrType = dicGetType();
			instrClass = dicGetClass();
			dicSetContext(CurrentContext);
		
			uc = yylex ();
			tab ();
			voidproc_affect ();
		}
		break;
	default:
		erreur ("unrecognized statement token");
	}
}

void
instructions (void)
{
	if (uc != BEND)
	{
		instr ();
		if (uc == COMMADOT)
		{
			uc = yylex ();
			instructions ();
		}
		else
			erreur ("missing ; in end of statement");
	}
	else
	{
		//Fin de bloc
		uc = yylex ();
	}
}

/*Section declaration*/
void
tab_proc (void)
{
	if (uc == PROCEDURE)
	{
		/*PCODE DEBUT */
		//memClass = CLASS_PROCEDURE ;
		erreur ("Unauthorized operation: a procedure cannot be passed as an argument");
		/*PCODE FIN */
		uc = yylex ();
	}
	else
	{
		if (uc == ARRAY)
		{
			memRef = 0;  //sinon: pointeur vers @=1
			memClass = CLASS_ARRAY;
			uc = yylex ();
		}
	}
}

void
specifier (void)
{
	/*PCODE DEBUT */
	//Specification des globales par défaut
	memClass = CLASS_VARIABLE;
	memType = SYM_INTEGER;
	/*PCODE FIN */
	switch (uc)
	{
    case STRING:
      erreur ("Unauthorized operation : STRING cannot be passed as argument");
      uc = yylex ();
    case ARRAY:
      memRef = 0;
      memClass = CLASS_ARRAY;
      uc = yylex ();
      break;
    case PROCEDURE:
      erreur ("Unauthorized operation: a procedure cannot be passed as an argument");
      break;
    case INTEGER:
    case REAL:
    case BOOLEAN:
      memRef=1;
      switch (uc)
      {
        case INTEGER:
          memType = SYM_INTEGER;
          break;
        case REAL:
          memType = SYM_REAL;
          break;
        case BOOLEAN:
          memType = SYM_BOOLEAN;
          break;
      }
      uc = yylex ();
      tab_proc ();
      break;
    default:
		erreur ("invalid specifier");
	}
}

void
spec (void)
{
	specifier ();
	//id_simple();
	/*Reconnaitre liste d'identifiants */
	if (uc == IDENT)
	{
		if (!dicSelect (yytext))
			erreur ("Unknown argument specified");

		dicSetType (memType);
		dicSetClass (memClass);
    
    if(memRef == 0 && memClass==CLASS_ARRAY)
      dicVoidArray(); //pas de bornes au tableau argument
    
		uc = yylex ();
		while (uc == COMMA)
		{
			uc = yylex ();
			if (uc != IDENT)
				erreur ("Missing IDENTIFIER after , in list of values");

			if (!dicCheckExisting (yytext))
				erreur ("Unknown argument specified");

			dicSelect (yytext);
			dicSetType (memType);
			dicSetClass (memClass);
      if(memRef == 0 && memClass==CLASS_ARRAY)
        dicVoidArray(); //pas de bornes au tableau argument
      
			uc = yylex ();
		}
		if (uc != COMMADOT)
			erreur ("Missing ; after Value specifier");
		uc = yylex ();
	}
	else
	{
		erreur ("Missing IDENTIFIER after value statement");
	}
}

static void ArgOnStack()
{
   int nArgs;
   dicSetContext(DIC_GLOBAL);
   nArgs= dicGetArgsQuantity();
   dicSetContext(DIC_LOCAL);

   if(nArgs > 0)
   {
      dicSelectFirst();
      AddVarBlock(dicGetName(),1);
      nArgs--;
      while(nArgs > 0)
      {
         dicSelectNext();
         AddVarBlock(dicGetName(),1);
         nArgs--;
      }
   }
}

void
specifications (void)
{
	switch (uc)
	{
	case STRING:
	case INTEGER:
	case REAL:
	case BOOLEAN:
	case PROCEDURE:
    memRef=1;
		spec ();
		specifications ();
		break;
  case ARRAY:
    memRef=0;
		spec ();
		specifications ();
    break;
	default: //cas avec specification
      //fin des specifications des arguments
      ArgOnStack();
		block (); 
		break;
	}
}

void
proc_specifiers (void)
{
	if (uc == VALUE)
	{
		uc = yylex ();
		//id_simple();
		/*Reconnaitre liste d'identifiants */
		if (uc == IDENT)
		{
			/*PCODE DEBUT */
			if (!dicCheckExisting (yytext))
				erreur ("Unknown argument specified");
			dicSelect (yytext);
			dicSetRef (ARG_VALUE);
			/*PCODE FIN */
			uc = yylex ();
			while (uc == COMMA)
			{
				uc = yylex ();
				if (uc != IDENT)
					erreur ("Missing IDENTIFIER after , in list of values");
				/*PCODE DEBUT */
				if (!dicCheckExisting (yytext))
					erreur ("Unknown argument specified");
				dicSelect (yytext);
				dicSetRef (ARG_VALUE);
				/*PCODE FIN */
				uc = yylex ();
			}
			if (uc != COMMADOT)
				erreur ("Missing ; after Value specifier");
			uc = yylex ();
		}
		else
		{
			erreur ("Missing IDENTIFIER after value statement");
		}
		specifications ();
	}
	else
	{
		switch (uc)
		{
		case STRING:
		case INTEGER:
		case REAL:
		case BOOLEAN:
		case PROCEDURE:
			specifications ();
			break;
    case ARRAY:
      specifications ();
			break;
		default:
         //aucune specification d'argument
         ArgOnStack();
			//Bloc
			block ();
		}
	}

	/*PCODE DEBUT */
	printf ("Local dico:\n");
	printLocalDic ();
  EndBlock ();
	dicSetContext (DIC_GLOBAL);
	dicCleanLocal ();
	ReturnProc (dicGetName());
	/*PCODE FIN */
}

void
id_suivantp (void)
{
	if (uc == COMMA)
	{
		uc = yylex ();
		id_list ();
	}
	else
	{
		if (uc != BRACKETR)
			erreur ("missing ) at the end of parameters list in procedure declaration");
		uc = yylex ();
	}
}

void
id_list (void)
{
	if (uc == IDENT)
	{
		/*PCODE DEBUT */
		dicSetContext (DIC_GLOBAL);
		dicIncArgsQuantity ();
		dicSetContext (DIC_LOCAL);
		dicAddEntry (yytext);
		/*PCODE FIN */
		uc = yylex ();
		id_suivantp ();
	}
	else
		erreur ("missing IDENTIFIER in list");
}

/*
void
param_externe (void)
{
	if (uc == IDENT)
	{
		uc = yylex ();
		if (uc == DOUBLEDOT)
		{
			uc = yylex ();
			if (uc == BRACKETL)
			{
				uc = yylex ();
				id_list ();
			}
		}
		else
			erreur ("Missing : for external parameter");
	}
	else
		erreur ("CHAIN OF CHARS missing for external parameter");
}

void
param_externes (void)
{
	while (uc != COMMADOT)
	{
		param_externe ();
	}
	uc = yylex ();
}
*/

void
id_list_vide (void)
{
	if (uc != BRACKETR)
	{
		id_list ();
	}
	else
		uc = yylex ();
}

void
proc_head (void)
{
	if (uc != BRACKETL)
		erreur ("missing ( after procedure identifier");
	uc = yylex ();
	/*PCODE DEBUT */
	//On est desormais dans un contexte local, si on y etait deja alors c'est une erreur
	if (DIC_LOCAL == dicGetContext ())
	{
		erreur ("Unauthorized operation: a procedure cannot declare another one");
	}
	
	dicSetContext (DIC_LOCAL);
	dicSetBuilderType (SYM_INTEGER);
	dicSetBuilderClass (CLASS_VARIABLE);
	dicSetBuilderRef (ARG_REFERENCE);

  CreateBlock();
	/*PCODE FIN */
	id_list_vide ();
	/*PCODE DEBUT */
	dicSetBuilderRef (ARG_VALUE);
	/*PCODE FIN */
	if (uc != COMMADOT)
		erreur ("Missing ; after header of procedure declaration");
	uc = yylex ();
	//param_externes();
}

void
id_suivant (void)
{
	if (uc == COMMA)
	{
		uc = yylex ();
		id_simple ();
	}
	else
	{
		if (uc != COMMADOT)
			erreur ("Termination ; missing after simple identifier declaration");
		uc = yylex ();
	}
}

void
id_simple (void)
{
	char *temp;
	if (uc == IDENT)
	{
		/*PCODE DEBUT */
		if (dicCheckExisting (yytext))
			erreur ("Unauthorized operation: variable already declared");
		dicAddEntry (yytext);
		temp = dicGetName();
		/*PCODE FIN */
		uc = yylex ();
		
		AddVarBlock(temp,0);
		
		id_suivant ();
	}
}

void
paire (void)
{
	if (uc == COMMA)
	{
		uc = yylex ();
		liste_bornes ();
	}
	else
	{
		if (uc != ARRAYBRACKETR)
			erreur ("missing ] at the end of extrema");
		uc = yylex ();
	}
}

void
paire_bornes (void)
{
	int valeurinf;
  int negate=1;
  
  if(uc == MINUS)
  {
    negate = -1;
    uc = yylex ();
  }
	if (uc != NBR_INT)
		erreur ("Missing constant integer value for table minimum");

	valeurinf = valeur * negate;
  negate =1;
	uc = yylex ();
  
	if (uc != DOUBLEDOT)
		erreur ("missing : for separator between table extrems");
	uc = yylex ();
  
	if (uc != NBR_INT)
		erreur ("Missing constant integer value for table maximum");

	dicSetArray (valeurinf, valeur*negate);
	uc = yylex ();
}

void
liste_bornes (void)
{
	paire_bornes ();
	paire ();
}

void
tableau_suivant (void)
{
	if (uc == COMMA)
	{
		uc = yylex ();
		id_tableau ();
	}
	else
	{
		if (uc != COMMADOT)
			erreur ("Termination ; missing at the end of table declaration");
		uc = yylex ();
	}
}

void
tableau (void)
{
	if (uc == ARRAY)
	{
		uc = yylex ();
		id_tableau ();
	}
	else
	{
		id_simple ();
	}
}

void
tableau_proc (void)
{
	switch (uc)
	{
	case PROCEDURE:
		/*PCODE DEBUT */
		dicSetBuilderClass (CLASS_PROCEDURE);
		/*PCODE FIN */
		uc = yylex ();
		proc ();
		break;
	case ARRAY:
		uc = yylex ();
		/*PCODE DEBUT */
		dicSetBuilderClass (CLASS_ARRAY);
		/*PCODE FIN */
		id_tableau ();
		break;
	default:
		id_simple ();
	}
}

void
type_var (void)
{
	switch (uc)
	{
	case ARRAY:
    dicSetBuilderType (SYM_INTEGER);
		uc = yylex ();
		id_tableau ();
		break;
	case INTEGER:
    dicSetBuilderType (SYM_INTEGER);
		uc = yylex ();
		tableau ();
		break;
	case BOOLEAN:
    dicSetBuilderType (SYM_BOOLEAN);
		uc = yylex ();
		tableau ();
		break;
	case REAL:
    dicSetBuilderType (SYM_REAL);
		uc = yylex ();
		tableau ();
		break;
	default:
		erreur ("unrecognized variable type");
	}
}

void
proc (void)
{
	if (uc != IDENT)
		erreur ("missing correct identifier for procedure declaration");
	/*PCODE DEBUT */
	dicAddEntry (yytext);

	CreateProc (yytext);

	/*PCODE FIN */
	uc = yylex ();
	proc_head ();
	proc_specifiers ();
}

void
id_tableau (void)
{
	char *temp;
	if (uc != IDENT)
		erreur ("missing IDENTIFIER for array declaration");
	/*PCODE DEBUT */
	dicAddEntry (yytext);
	temp = dicGetName();
	/*PCODE FIN */
	uc = yylex ();
	if (uc != ARRAYBRACKETL)
		erreur ("missing [ after IDENTIFIER in array declaration");
	uc = yylex ();
	liste_bornes ();
	
	AddVarBlock(temp,0);
	
	tableau_suivant ();
}

void
declaration (void)
{
	//Une seule declaration pouvant concerner plusieurs variables
	//Specialisation par type declare  /*PCODE DEBUT*/
	/*Initialiser le constructeur par defaut */
	dicSetBuilderType (SYM_INTEGER);
	dicSetBuilderClass (CLASS_VARIABLE);
	dicSetBuilderOffset (0);
	/*PCODE FIN */
	switch (uc)
	{
	case ARRAY:
		/*PCODE DEBUT */
		dicSetBuilderClass (CLASS_ARRAY);
		/*PCODE FIN */
		uc = yylex ();
		id_tableau ();
		break;
	case INTEGER:
	case REAL:
	case BOOLEAN:
		/*PCODE DEBUT */
		switch (uc)
		{
		case BOOLEAN:
			dicSetBuilderType (SYM_BOOLEAN);
			break;
		case REAL:
			dicSetBuilderType (SYM_REAL);
			break;
		}
		/*PCODE FIN */
		uc = yylex ();
		tableau_proc ();
		break;
	case OWN:
		/*PCODE DEBUT */
		dicSetBuilderOffset (-1);
		/*PCODE FIN */
		uc = yylex ();
		type_var ();
		break;
	case PROCEDURE:
		/*PCODE DEBUT */
		dicSetBuilderType (SYM_VOID);
		dicSetBuilderClass (CLASS_PROCEDURE);
		/*PCODE FIN */
		uc = yylex ();
		proc ();
		break;
	}
}

void
instr_unique ()
{
	instr ();
}

void
block (void)
{
	//On attend un BEGIN
	if (uc != BBEGIN)
		erreur ("missing 'BEGIN' at beginning of block!");
	//On avance
	uc = yylex ();
	//Ensemble de Declarations optionnel
	while (uc == INTEGER || uc == OWN || uc == REAL || uc == BOOLEAN
	       || uc == PROCEDURE || uc == ARRAY)
	{
		declaration ();
	}
	//Ensemble d'instructions optionnelles
	while (uc == IDENT || uc == FOR || uc == IF)
	{
		instr ();
	}
	if (uc != BEND)
		erreur ("missing 'END' at the end of block!");
	uc = yylex ();
}

void
body (void)
{
	if (uc == BBEGIN)
	{
		block ();
	}
	else
	{
		instr_unique ();
	}
}

void
program (void)
{
  CreateBlock();
	body ();
	EndBlock ();
}

int
main (int argc, char *argv[])
{
	int i;
	FILE *f;
	char *nomfic;
	char command[256];
	if (argc != 1 + 1)
	{
		fprintf (stderr, "USAGE: %s source_file.alg\n",
			 argv[0]);
		exit (EXIT_FAILURE);
	}
	yyin = fopen (argv[1], "r");
	uc = yylex ();
  	CreateFileText("text.s");
	CreateFileData("data.s");
	program ();
	printf ("ALGOL60 syntax was successfully recognized\n");
	fclose (yyin);
	CloseFileData();
  	CloseFileText();
	printf ("global dico:\n");
	printGlobalDic ();
	return EXIT_SUCCESS;
}
